@@ -1,4 +1,5 @@
 use std::collections::HashSet;
+use std::collections::VecDeque;
 use std::io::Write;
 
 use bstr::ByteSlice;
@@ -19,6 +20,9 @@ struct State {
     fixup: git_stack::config::Fixup,
     dry_run: bool,
     snapshot_capacity: Option<usize>,
+    protect_commit_count: Option<usize>,
+    protect_commit_age: std::time::Duration,
+    protect_commit_time: std::time::SystemTime,
 
     show_format: git_stack::config::Format,
     show_stacked: bool,
@@ -63,7 +67,9 @@ impl State {
         .with_code(proc_exit::Code::CONFIG_ERR)?;
         let dry_run = args.dry_run;
         let snapshot_capacity = repo_config.capacity();
-
+        let protect_commit_count = repo_config.protect_commit_count();
+        let protect_commit_age = repo_config.protect_commit_age();
+        let protect_commit_time = std::time::SystemTime::now() - protect_commit_age;
         let show_format = repo_config.show_format();
         let show_stacked = repo_config.show_stacked();
 
@@ -176,6 +182,9 @@ impl State {
             fixup,
             dry_run,
             snapshot_capacity,
+            protect_commit_count,
+            protect_commit_age,
+            protect_commit_time,
 
             show_format,
             show_stacked,
@@ -225,7 +234,11 @@ impl StackState {
     }
 }
 
-pub fn stack(args: &crate::args::Args, colored_stdout: bool) -> proc_exit::ExitResult {
+pub fn stack(
+    args: &crate::args::Args,
+    colored_stdout: bool,
+    colored_stderr: bool,
+) -> proc_exit::ExitResult {
     log::trace!("Initializing");
     let cwd = std::env::current_dir().with_code(proc_exit::Code::USAGE_ERR)?;
     let repo = git2::Repository::discover(&cwd).with_code(proc_exit::Code::USAGE_ERR)?;
@@ -243,10 +256,21 @@ pub fn stack(args: &crate::args::Args, colored_stdout: bool) -> proc_exit::ExitR
         }
 
         // Update status of remote unprotected branches
-        match git_fetch(&mut state.repo) {
-            Ok(_) => (),
-            Err(err) => {
-                log::warn!("Skipping fetch of `{}`, {}", state.repo.push_remote(), err);
+        let mut push_branches: Vec<_> = state
+            .stacks
+            .iter()
+            .flat_map(|stack| stack.branches.iter())
+            .filter(|(oid, _)| !state.protected_branches.contains_oid(*oid))
+            .flat_map(|(_, b)| b.iter())
+            .filter_map(|b| b.push_id.map(|_| b.name.as_str()))
+            .collect();
+        push_branches.sort_unstable();
+        if !push_branches.is_empty() {
+            match git_fetch(&mut state.repo, &push_branches, state.dry_run) {
+                Ok(_) => (),
+                Err(err) => {
+                    log::warn!("Skipping fetch of `{}`, {}", state.repo.push_remote(), err);
+                }
             }
         }
 
@@ -360,7 +384,7 @@ pub fn stack(args: &crate::args::Args, colored_stdout: bool) -> proc_exit::ExitR
         state.update().with_code(proc_exit::Code::FAILURE)?;
     }
 
-    show(&state, colored_stdout).with_code(proc_exit::Code::FAILURE)?;
+    show(&state, colored_stdout, colored_stderr).with_code(proc_exit::Code::FAILURE)?;
 
     git_stack::git::stash_pop(&mut state.repo, stash_id);
 
@@ -381,17 +405,24 @@ fn plan_changes(state: &State, stack: &StackState) -> eyre::Result<git_stack::gi
         .repo
         .find_commit(stack.base.id)
         .expect("base branch is valid");
-    let root = git_stack::graph::Node::from_branches(&state.repo, graphed_branches)?;
-    let mut root = root.extend(&state.repo, git_stack::graph::Node::new(base_commit))?;
-    git_stack::graph::protect_branches(&mut root, &state.repo, &state.protected_branches);
+    let mut graph = git_stack::graph::Graph::from_branches(&state.repo, graphed_branches)?;
+    graph.insert(&state.repo, git_stack::graph::Node::new(base_commit))?;
+    git_stack::graph::protect_branches(&mut graph, &state.repo, &state.protected_branches);
+    if let Some(protect_commit_count) = state.protect_commit_count {
+        git_stack::graph::protect_large_branches(&mut graph, protect_commit_count);
+    }
+    git_stack::graph::protect_old_branches(&mut graph, state.protect_commit_time);
+    if let Some(user) = state.repo.user() {
+        git_stack::graph::protect_foreign_branches(&mut graph, &user);
+    }
 
     if state.rebase {
-        git_stack::graph::rebase_branches(&mut root, stack.onto.id);
-        git_stack::graph::drop_by_tree_id(&mut root);
+        git_stack::graph::rebase_branches(&mut graph, stack.onto.id);
+        git_stack::graph::drop_by_tree_id(&mut graph);
     }
-    git_stack::graph::fixup(&mut root, state.fixup);
+    git_stack::graph::fixup(&mut graph, state.fixup);
 
-    let script = git_stack::graph::to_script(&root);
+    let script = git_stack::graph::to_script(&graph);
 
     Ok(script)
 }
@@ -402,73 +433,144 @@ fn push(state: &mut State) -> eyre::Result<()> {
         let stack_graphed_branches = stack.graphed_branches();
         graphed_branches.extend(stack_graphed_branches.into_iter().flat_map(|(_, b)| b));
     }
-    let root = git_stack::graph::Node::from_branches(&state.repo, graphed_branches)?;
-    let mut root = root.extend(
+    let mut graph = git_stack::graph::Graph::from_branches(&state.repo, graphed_branches)?;
+    graph.insert(
         &state.repo,
         git_stack::graph::Node::new(state.head_commit.clone()),
     )?;
 
-    git_stack::graph::protect_branches(&mut root, &state.repo, &state.protected_branches);
-    git_stack::graph::pushable(&mut root);
+    git_stack::graph::protect_branches(&mut graph, &state.repo, &state.protected_branches);
+    if let Some(protect_commit_count) = state.protect_commit_count {
+        git_stack::graph::protect_large_branches(&mut graph, protect_commit_count);
+    }
+    git_stack::graph::protect_old_branches(&mut graph, state.protect_commit_time);
+    if let Some(user) = state.repo.user() {
+        git_stack::graph::protect_foreign_branches(&mut graph, &user);
+    }
 
-    git_push(&mut state.repo, &root, state.dry_run)?;
+    git_stack::graph::pushable(&mut graph);
+
+    git_push(&mut state.repo, &graph, state.dry_run)?;
 
     Ok(())
 }
 
-fn show(state: &State, colored_stdout: bool) -> eyre::Result<()> {
-    let mut roots = state
-        .stacks
-        .iter()
-        .map(|stack| -> eyre::Result<git_stack::graph::Node> {
-            let graphed_branches = stack.graphed_branches();
-            let base_commit = state
-                .repo
-                .find_commit(stack.base.id)
-                .expect("base branch is valid");
-            let root = git_stack::graph::Node::from_branches(&state.repo, graphed_branches)?;
-            let mut root = root.extend(&state.repo, git_stack::graph::Node::new(base_commit))?;
-            git_stack::graph::protect_branches(&mut root, &state.repo, &state.protected_branches);
+fn show(state: &State, colored_stdout: bool, colored_stderr: bool) -> eyre::Result<()> {
+    let palette_stderr = if colored_stderr {
+        Palette::colored()
+    } else {
+        Palette::plain()
+    };
+    let mut empty_stacks = Vec::new();
+    let mut old_stacks = Vec::new();
+    let mut foreign_stacks = Vec::new();
 
-            if state.dry_run {
-                // Show as-if we performed all mutations
-                if state.rebase {
-                    git_stack::graph::rebase_branches(&mut root, stack.onto.id);
-                    git_stack::graph::drop_by_tree_id(&mut root);
-                }
-                git_stack::graph::fixup(&mut root, state.fixup);
+    let mut graphs = Vec::with_capacity(state.stacks.len());
+    for stack in state.stacks.iter() {
+        let graphed_branches = stack.graphed_branches();
+        if graphed_branches.len() == 1 {
+            let branches = graphed_branches.iter().next().unwrap().1;
+            if branches.len() == 1 && branches[0].id != state.head_commit.id {
+                empty_stacks.push(format!("{}", palette_stderr.info.paint(&branches[0].name)));
+                continue;
             }
+        }
 
-            eyre::Result::Ok(root)
-        });
-    let mut root = roots.next().unwrap_or_else(|| {
-        let root = git_stack::graph::Node::new(state.head_commit.clone());
-        Ok(root)
-    })?;
-    for other in roots {
-        root = root.extend(&state.repo, other?)?;
+        let base_commit = state
+            .repo
+            .find_commit(stack.base.id)
+            .expect("base branch is valid");
+        let mut graph = git_stack::graph::Graph::from_branches(&state.repo, graphed_branches)?;
+        graph.insert(&state.repo, git_stack::graph::Node::new(base_commit))?;
+        git_stack::graph::protect_branches(&mut graph, &state.repo, &state.protected_branches);
+        if let Some(protect_commit_count) = state.protect_commit_count {
+            let protected =
+                git_stack::graph::protect_large_branches(&mut graph, protect_commit_count);
+            if !protected.is_empty() {
+                log::warn!(
+                    "Branches contain more than {} commits (should these be protected?): {}",
+                    protect_commit_count,
+                    protected.join("m ")
+                );
+            }
+        }
+        old_stacks.extend(
+            git_stack::graph::trim_old_branches(
+                &mut graph,
+                state.protect_commit_time,
+                &[state.head_commit.id],
+            )
+            .into_iter()
+            .map(|b| format!("{}", palette_stderr.warn.paint(b))),
+        );
+        if let Some(user) = state.repo.user() {
+            foreign_stacks.extend(
+                git_stack::graph::trim_foreign_branches(&mut graph, &user, &[state.head_commit.id])
+                    .into_iter()
+                    .map(|b| format!("{}", palette_stderr.warn.paint(b))),
+            );
+        }
+
+        if state.dry_run {
+            // Show as-if we performed all mutations
+            if state.rebase {
+                git_stack::graph::rebase_branches(&mut graph, stack.onto.id);
+                git_stack::graph::drop_by_tree_id(&mut graph);
+            }
+            git_stack::graph::fixup(&mut graph, state.fixup);
+        }
+
+        git_stack::graph::pushable(&mut graph);
+
+        graphs.push(graph);
+    }
+    if graphs.is_empty() {
+        let graph =
+            git_stack::graph::Graph::new(git_stack::graph::Node::new(state.head_commit.clone()));
+        graphs.push(graph);
+    }
+    graphs.sort_by_key(|g| {
+        let mut revwalk = state.repo.raw().revwalk().unwrap();
+        // Reduce the number of commits to walk
+        revwalk.simplify_first_parent().unwrap();
+        revwalk.push(g.root_id()).unwrap();
+        revwalk.count()
+    });
+
+    for graph in graphs {
+        match state.show_format {
+            git_stack::config::Format::Silent => (),
+            git_stack::config::Format::Branches
+            | git_stack::config::Format::BranchCommits
+            | git_stack::config::Format::Commits => {
+                write!(
+                    std::io::stdout(),
+                    "{}",
+                    DisplayTree::new(&state.repo, &graph)
+                        .colored(colored_stdout)
+                        .show(state.show_format)
+                        .stacked(state.show_stacked)
+                        .protected_branches(&state.protected_branches)
+                )?;
+            }
+            git_stack::config::Format::Debug => {
+                writeln!(std::io::stdout(), "{:#?}", graph)?;
+            }
+        }
     }
 
-    git_stack::graph::pushable(&mut root);
-
-    match state.show_format {
-        git_stack::config::Format::Silent => (),
-        git_stack::config::Format::Branches
-        | git_stack::config::Format::BranchCommits
-        | git_stack::config::Format::Commits => {
-            writeln!(
-                std::io::stdout(),
-                "{}",
-                DisplayTree::new(&state.repo, &root)
-                    .colored(colored_stdout)
-                    .show(state.show_format)
-                    .stacked(state.show_stacked)
-                    .protected_branches(&state.protected_branches)
-            )?;
-        }
-        git_stack::config::Format::Debug => {
-            writeln!(std::io::stdout(), "{:#?}", root)?;
-        }
+    if !empty_stacks.is_empty() {
+        log::info!("Empty stacks: {}", empty_stacks.join(", "));
+    }
+    if !old_stacks.is_empty() {
+        log::info!(
+            "Stacks older than {}: {}",
+            humantime::format_duration(state.protect_commit_age),
+            old_stacks.join(", ")
+        );
+    }
+    if !foreign_stacks.is_empty() {
+        log::info!("Stack from other users: {}", foreign_stacks.join(", "));
     }
 
     Ok(())
@@ -508,18 +610,57 @@ fn resolve_implicit_base(
     Ok(branch.clone())
 }
 
-fn git_fetch(repo: &mut git_stack::git::GitRepo) -> eyre::Result<()> {
+fn git_fetch(
+    repo: &mut git_stack::git::GitRepo,
+    branches: &[&str],
+    dry_run: bool,
+) -> eyre::Result<()> {
     let remote = repo.push_remote();
-    log::debug!("git fetch {}", remote);
+    log::debug!("git fetch {} {}", remote, branches.join(" "));
+    if dry_run {
+        return Ok(());
+    }
+
+    let output = std::process::Command::new("git")
+        .arg("ls-remote")
+        .arg("--heads")
+        .arg(remote)
+        .args(branches)
+        .stdout(std::process::Stdio::piped())
+        .spawn()
+        .wrap_err("Could not run `git fetch`")?
+        .wait_with_output()?;
+    if !output.status.success() {
+        eyre::bail!("Could not run `git fetch`");
+    }
+    let stdout = String::from_utf8(output.stdout).wrap_err("Could not run `git fetch`")?;
+    let remote_branches: Vec<_> = stdout
+        .lines()
+        .filter_map(|l| l.rsplit_once('/'))
+        .map(|s| s.1)
+        .collect();
+
+    for branch in branches {
+        if !remote_branches.contains(branch) {
+            let remote_branch = format!("{}/{}", remote, branch);
+            log::info!("Pruning {}", remote_branch);
+            let mut branch = repo
+                .raw()
+                .find_branch(&remote_branch, git2::BranchType::Remote)?;
+            branch.delete()?;
+        }
+    }
+
     // A little uncertain about some of the weirder authentication needs, just deferring to `git`
     // instead of using `libgit2`
     let status = std::process::Command::new("git")
         .arg("fetch")
         .arg(remote)
+        .args(remote_branches)
         .status()
         .wrap_err("Could not run `git fetch`")?;
     if !status.success() {
-        eyre::bail!("`git fetch {}` failed", remote);
+        eyre::bail!("`git fetch {} {}` failed", remote, branches.join(" "));
     }
 
     Ok(())
@@ -756,10 +897,23 @@ fn drop_branches(
 
 fn git_push(
     repo: &mut git_stack::git::GitRepo,
-    node: &git_stack::graph::Node,
+    graph: &git_stack::graph::Graph,
     dry_run: bool,
 ) -> eyre::Result<()> {
-    let failed = git_push_internal(repo, node, dry_run);
+    let mut failed = Vec::new();
+
+    let mut node_queue = VecDeque::new();
+    node_queue.push_back(graph.root_id());
+    while let Some(current_id) = node_queue.pop_front() {
+        let current = graph.get(current_id).expect("all children exist");
+
+        failed.extend(git_push_node(repo, current, dry_run));
+
+        for child_id in current.children.iter().copied() {
+            node_queue.push_back(child_id);
+        }
+    }
+
     if failed.is_empty() {
         Ok(())
     } else {
@@ -767,7 +921,7 @@ fn git_push(
     }
 }
 
-fn git_push_internal(
+fn git_push_node(
     repo: &mut git_stack::git::GitRepo,
     node: &git_stack::graph::Node,
     dry_run: bool,
@@ -808,18 +962,12 @@ fn git_push_internal(
         }
     }
 
-    if failed.is_empty() {
-        for child in node.children.values() {
-            failed.extend(git_push_internal(repo, child, dry_run));
-        }
-    }
-
     failed
 }
 
 struct DisplayTree<'r> {
     repo: &'r git_stack::git::GitRepo,
-    root: &'r git_stack::graph::Node,
+    graph: &'r git_stack::graph::Graph,
     protected_branches: git_stack::git::Branches,
     palette: Palette,
     show: git_stack::config::Format,
@@ -827,10 +975,10 @@ struct DisplayTree<'r> {
 }
 
 impl<'r> DisplayTree<'r> {
-    pub fn new(repo: &'r git_stack::git::GitRepo, root: &'r git_stack::graph::Node) -> Self {
+    pub fn new(repo: &'r git_stack::git::GitRepo, graph: &'r git_stack::graph::Graph) -> Self {
         Self {
             repo,
-            root,
+            graph,
             protected_branches: Default::default(),
             palette: Palette::plain(),
             show: Default::default(),
@@ -866,40 +1014,40 @@ impl<'r> DisplayTree<'r> {
 impl<'r> std::fmt::Display for DisplayTree<'r> {
     fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> Result<(), std::fmt::Error> {
         let head_branch = self.repo.head_branch().unwrap();
+
+        let is_visible: Box<dyn Fn(&git_stack::graph::Node) -> bool> = match self.show {
+            git_stack::config::Format::Silent => unreachable!("No silent view for tree"),
+            git_stack::config::Format::Commits => Box::new(|_| false),
+            git_stack::config::Format::BranchCommits => Box::new(|node| {
+                let interesting_commit =
+                    node.commit.id == head_branch.id || node.commit.id == self.graph.root_id();
+                let boring_commit = node.branches.is_empty() && node.children.len() == 1;
+                let protected = node.action.is_protected();
+                interesting_commit || !boring_commit || !protected
+            }),
+            git_stack::config::Format::Branches => Box::new(|node| {
+                let interesting_commit =
+                    node.commit.id == head_branch.id || node.commit.id == self.graph.root_id();
+                let boring_commit = node.branches.is_empty() && node.children.len() == 1;
+                interesting_commit || !boring_commit
+            }),
+            git_stack::config::Format::Debug => unreachable!("No debug view for tree"),
+        };
+
         let mut tree = to_tree(
             self.repo,
             &head_branch,
             &self.protected_branches,
-            self.root,
+            self.graph,
+            self.graph.root_id(),
             &self.palette,
+            &is_visible,
         );
         if self.stacked {
             tree.linearize();
         } else {
             tree.sort();
         }
-        match self.show {
-            git_stack::config::Format::Silent => tree.skip(|_| true),
-            git_stack::config::Format::Commits => tree.skip(|_| false),
-            git_stack::config::Format::BranchCommits => tree.skip(|tree| {
-                if let Some(node) = tree.root.node {
-                    let protected = node.action.is_protected();
-                    let boring_commit = node.branches.is_empty() && tree.stacks.is_empty();
-                    protected && boring_commit
-                } else {
-                    false
-                }
-            }),
-            git_stack::config::Format::Branches => tree.skip(|tree| {
-                if let Some(node) = tree.root.node {
-                    let boring_commit = node.branches.is_empty() && tree.stacks.is_empty();
-                    boring_commit
-                } else {
-                    false
-                }
-            }),
-            git_stack::config::Format::Debug => tree.skip(|_| false),
-        }
         let tree = tree.into_display();
         tree.fmt(f)
     }
@@ -909,37 +1057,55 @@ fn to_tree<'r>(
     repo: &'r git_stack::git::GitRepo,
     head_branch: &'r git_stack::git::Branch,
     protected_branches: &'r git_stack::git::Branches,
-    node: &'r git_stack::graph::Node,
+    graph: &'r git_stack::graph::Graph,
+    mut node_id: git2::Oid,
     palette: &'r Palette,
+    is_visible: &dyn Fn(&git_stack::graph::Node) -> bool,
 ) -> Tree<'r> {
-    let mut weight = if node.action.is_protected() {
-        Weight::Protected(0)
-    } else if node.commit.id == head_branch.id {
-        Weight::Head(0)
-    } else {
-        Weight::Commit(0)
-    };
+    loop {
+        let node = graph.get(node_id).expect("all children exist");
+        // The API requires us to handle 0 or many children, so not checking visibility
+        if node.children.len() == 1 && !is_visible(node) {
+            node_id = node.children.iter().copied().next().unwrap();
+            continue;
+        }
 
-    let mut stacks = Vec::new();
-    for child in node.children.values() {
-        let child_tree = to_tree(repo, head_branch, protected_branches, child, palette);
-        weight = weight.max(child_tree.weight);
-        stacks.push(vec![child_tree]);
+        let mut weight = if node.action.is_protected() {
+            Weight::Protected(0)
+        } else if node.commit.id == head_branch.id {
+            Weight::Head(0)
+        } else {
+            Weight::Commit(0)
+        };
+
+        let mut stacks = Vec::new();
+        for child_id in node.children.iter().copied() {
+            let child_tree = to_tree(
+                repo,
+                head_branch,
+                protected_branches,
+                graph,
+                child_id,
+                palette,
+                is_visible,
+            );
+            weight = weight.max(child_tree.weight);
+            stacks.push(vec![child_tree]);
+        }
+
+        let tree = Tree {
+            root: RenderNode {
+                repo,
+                head_branch,
+                protected_branches,
+                node: Some(node),
+                palette,
+            },
+            weight,
+            stacks,
+        };
+        return tree;
     }
-
-    let tree = Tree {
-        root: RenderNode {
-            repo,
-            head_branch,
-            protected_branches,
-            node: Some(node),
-            palette,
-        },
-        weight,
-        stacks,
-    };
-
-    tree
 }
 
 struct Tree<'r> {
@@ -949,26 +1115,6 @@ struct Tree<'r> {
 }
 
 impl<'r> Tree<'r> {
-    fn skip<F>(&mut self, is_skipped: F)
-    where
-        F: Fn(&Self) -> bool + Copy,
-    {
-        for stack in self.stacks.iter_mut() {
-            let mut skipped = Vec::new();
-            for (index, child) in stack.iter_mut().enumerate() {
-                if is_skipped(child) {
-                    skipped.push(index)
-                } else {
-                    child.skip(is_skipped);
-                }
-            }
-            skipped.reverse();
-            for index in skipped {
-                stack.remove(index);
-            }
-        }
-    }
-
     fn sort(&mut self) {
         self.stacks.sort_by_key(|s| s[0].weight);
         for stack in self.stacks.iter_mut() {
