@@ -1,9 +1,10 @@
 use std::str::FromStr;
 
-#[derive(Default, Clone, Debug, serde::Serialize, serde::Deserialize)]
-#[serde(rename_all = "kebab-case")]
+#[derive(Default, Clone, Debug)]
 pub struct RepoConfig {
     pub protected_branches: Option<Vec<String>>,
+    pub protect_commit_count: Option<usize>,
+    pub protect_commit_age: Option<std::time::Duration>,
     pub stack: Option<Stack>,
     pub push_remote: Option<String>,
     pub pull_remote: Option<String>,
@@ -15,6 +16,8 @@ pub struct RepoConfig {
 }
 
 static PROTECTED_STACK_FIELD: &str = "stack.protected-branch";
+static PROTECT_COMMIT_COUNT: &str = "stack.protect-commit-count";
+static PROTECT_COMMIT_AGE: &str = "stack.protect-commit-age";
 static STACK_FIELD: &str = "stack.stack";
 static PUSH_REMOTE_FIELD: &str = "stack.push-remote";
 static PULL_REMOTE_FIELD: &str = "stack.pull-remote";
@@ -24,6 +27,9 @@ static AUTO_FIXUP_FIELD: &str = "stack.auto-fixup";
 static BACKUP_CAPACITY_FIELD: &str = "branch-stash.capacity";
 
 static DEFAULT_PROTECTED_BRANCHES: [&str; 4] = ["main", "master", "dev", "stable"];
+static DEFAULT_PROTECT_COMMIT_COUNT: usize = 50;
+static DEFAULT_PROTECT_COMMIT_AGE: std::time::Duration =
+    std::time::Duration::from_secs(60 * 60 * 24 * 14);
 const DEFAULT_CAPACITY: usize = 30;
 
 impl RepoConfig {
@@ -111,6 +117,17 @@ impl RepoConfig {
                         .get_or_insert_with(Vec::new)
                         .push(value.into_owned());
                 }
+            } else if key == PROTECT_COMMIT_COUNT {
+                if let Some(value) = value.as_ref().and_then(|v| FromStr::from_str(v).ok()) {
+                    config.protect_commit_count = Some(value);
+                }
+            } else if key == PROTECT_COMMIT_AGE {
+                if let Some(value) = value
+                    .as_ref()
+                    .and_then(|v| humantime::parse_duration(v).ok())
+                {
+                    config.protect_commit_age = Some(value);
+                }
             } else if key == STACK_FIELD {
                 if let Some(value) = value.as_ref().and_then(|v| FromStr::from_str(v).ok()) {
                     config.stack = Some(value);
@@ -161,11 +178,14 @@ impl RepoConfig {
 
     fn from_defaults_internal(config: Option<&git2::Config>) -> Self {
         let mut conf = Self::default();
+        conf.protect_commit_count = Some(conf.protect_commit_count().unwrap_or(0));
+        conf.protect_commit_age = Some(conf.protect_commit_age());
         conf.stack = Some(conf.stack());
         conf.push_remote = Some(conf.push_remote().to_owned());
         conf.pull_remote = Some(conf.pull_remote().to_owned());
         conf.show_format = Some(conf.show_format());
         conf.show_stacked = Some(conf.show_stacked());
+        conf.auto_fixup = Some(conf.auto_fixup());
         conf.capacity = Some(DEFAULT_CAPACITY);
 
         let mut protected_branches: Vec<String> = Vec::new();
@@ -200,25 +220,34 @@ impl RepoConfig {
             })
             .unwrap_or(None);
 
+        let protect_commit_count = config
+            .get_i64(PROTECT_COMMIT_COUNT)
+            .ok()
+            .map(|i| i.max(0) as usize);
+        let protect_commit_age = config
+            .get_string(AUTO_FIXUP_FIELD)
+            .ok()
+            .and_then(|s| humantime::parse_duration(&s).ok());
+
         let push_remote = config.get_string(PUSH_REMOTE_FIELD).ok();
         let pull_remote = config.get_string(PULL_REMOTE_FIELD).ok();
 
         let stack = config
-            .get_str(STACK_FIELD)
+            .get_string(STACK_FIELD)
             .ok()
-            .and_then(|s| FromStr::from_str(s).ok());
+            .and_then(|s| FromStr::from_str(&s).ok());
 
         let show_format = config
-            .get_str(FORMAT_FIELD)
+            .get_string(FORMAT_FIELD)
             .ok()
-            .and_then(|s| FromStr::from_str(s).ok());
+            .and_then(|s| FromStr::from_str(&s).ok());
 
         let show_stacked = config.get_bool(STACKED_FIELD).ok();
 
         let auto_fixup = config
-            .get_str(AUTO_FIXUP_FIELD)
+            .get_string(AUTO_FIXUP_FIELD)
             .ok()
-            .and_then(|s| FromStr::from_str(s).ok());
+            .and_then(|s| FromStr::from_str(&s).ok());
 
         let capacity = config
             .get_i64(BACKUP_CAPACITY_FIELD)
@@ -227,6 +256,8 @@ impl RepoConfig {
 
         Self {
             protected_branches,
+            protect_commit_count,
+            protect_commit_age,
             push_remote,
             pull_remote,
             stack,
@@ -264,7 +295,8 @@ impl RepoConfig {
             (None, Some(rhs)) => self.protected_branches = Some(rhs),
             (_, _) => (),
         }
-
+        self.protect_commit_count = other.protect_commit_count.or(self.protect_commit_count);
+        self.protect_commit_age = other.protect_commit_age.or(self.protect_commit_age);
         self.push_remote = other.push_remote.or(self.push_remote);
         self.pull_remote = other.pull_remote.or(self.pull_remote);
         self.stack = other.stack.or(self.stack);
@@ -280,6 +312,18 @@ impl RepoConfig {
         self.protected_branches.as_deref().unwrap_or(&[])
     }
 
+    pub fn protect_commit_count(&self) -> Option<usize> {
+        let protect_commit_count = self
+            .protect_commit_count
+            .unwrap_or(DEFAULT_PROTECT_COMMIT_COUNT);
+        (protect_commit_count != 0).then(|| protect_commit_count)
+    }
+
+    pub fn protect_commit_age(&self) -> std::time::Duration {
+        self.protect_commit_age
+            .unwrap_or(DEFAULT_PROTECT_COMMIT_AGE)
+    }
+
     pub fn push_remote(&self) -> &str {
         self.push_remote.as_deref().unwrap_or("origin")
     }
@@ -323,6 +367,18 @@ impl std::fmt::Display for RepoConfig {
                 branch
             )?;
         }
+        writeln!(
+            f,
+            "\t{}={}",
+            PROTECT_COMMIT_COUNT.split_once(".").unwrap().1,
+            self.protect_commit_count().unwrap_or(0)
+        )?;
+        writeln!(
+            f,
+            "\t{}={}",
+            PROTECT_COMMIT_AGE.split_once(".").unwrap().1,
+            humantime::format_duration(self.protect_commit_age())
+        )?;
         writeln!(
             f,
             "\t{}={}",
@@ -378,8 +434,7 @@ fn default_branch(config: &git2::Config) -> &str {
     config.get_str("init.defaultBranch").ok().unwrap_or("main")
 }
 
-#[derive(Debug, Copy, Clone, PartialEq, Eq, serde::Serialize, serde::Deserialize)]
-#[serde(rename_all = "kebab-case")]
+#[derive(Debug, Copy, Clone, PartialEq, Eq)]
 pub enum Format {
     Silent,
     Branches,
@@ -426,8 +481,7 @@ impl Default for Format {
     }
 }
 
-#[derive(Debug, Copy, Clone, PartialEq, Eq, serde::Serialize, serde::Deserialize)]
-#[serde(rename_all = "kebab-case")]
+#[derive(Debug, Copy, Clone, PartialEq, Eq)]
 pub enum Stack {
     Current,
     Dependents,
@@ -471,8 +525,7 @@ impl Default for Stack {
     }
 }
 
-#[derive(Debug, Copy, Clone, PartialEq, Eq, serde::Serialize, serde::Deserialize)]
-#[serde(rename_all = "kebab-case")]
+#[derive(Debug, Copy, Clone, PartialEq, Eq)]
 pub enum Fixup {
     Ignore,
     Move,
