@@ -2,6 +2,8 @@ use bstr::ByteSlice;
 use itertools::Itertools;
 
 pub trait Repo {
+    fn user(&self) -> Option<std::rc::Rc<str>>;
+
     fn is_dirty(&self) -> bool;
     fn merge_base(&self, one: git2::Oid, two: git2::Oid) -> Option<git2::Oid>;
 
@@ -49,6 +51,9 @@ pub struct Commit {
     pub id: git2::Oid,
     pub tree_id: git2::Oid,
     pub summary: bstr::BString,
+    pub time: std::time::SystemTime,
+    pub author: Option<std::rc::Rc<str>>,
+    pub committer: Option<std::rc::Rc<str>>,
 }
 
 impl Commit {
@@ -94,15 +99,17 @@ pub struct GitRepo {
     push_remote: Option<String>,
     pull_remote: Option<String>,
     commits: std::cell::RefCell<std::collections::HashMap<git2::Oid, std::rc::Rc<Commit>>>,
+    interned_strings: std::cell::RefCell<std::collections::HashSet<std::rc::Rc<str>>>,
 }
 
 impl GitRepo {
     pub fn new(repo: git2::Repository) -> Self {
         Self {
             repo,
-            commits: Default::default(),
             push_remote: None,
             pull_remote: None,
+            commits: Default::default(),
+            interned_strings: Default::default(),
         }
     }
 
@@ -126,6 +133,13 @@ impl GitRepo {
         &self.repo
     }
 
+    pub fn user(&self) -> Option<std::rc::Rc<str>> {
+        self.repo
+            .signature()
+            .ok()
+            .and_then(|s| s.name().map(|n| self.intern_string(n)))
+    }
+
     pub fn is_dirty(&self) -> bool {
         if self.repo.state() != git2::RepositoryState::Clean {
             log::trace!("Repository status is unclean: {:?}", self.repo.state());
@@ -161,10 +175,18 @@ impl GitRepo {
         } else {
             let commit = self.repo.find_commit(id).ok()?;
             let summary: bstr::BString = commit.summary_bytes().unwrap().into();
+            let time = std::time::SystemTime::UNIX_EPOCH
+                + std::time::Duration::from_secs(commit.time().seconds().max(0) as u64);
+
+            let author = commit.author().name().map(|n| self.intern_string(n));
+            let committer = commit.author().name().map(|n| self.intern_string(n));
             let commit = std::rc::Rc::new(Commit {
                 id: commit.id(),
                 tree_id: commit.tree_id(),
                 summary,
+                time,
+                author,
+                committer,
             });
             commits.insert(id, std::rc::Rc::clone(&commit));
             Some(commit)
@@ -367,13 +389,16 @@ impl GitRepo {
         into_id: git2::Oid,
     ) -> Result<git2::Oid, git2::Error> {
         // Based on https://www.pygit2.org/recipes/git-cherry-pick.html
-        let base_id = self.repo.merge_base(head_id, into_id)?;
-        let base_commit = self.repo.find_commit(base_id)?;
-        let base_tree = self.repo.find_tree(base_commit.tree_id())?;
-
         let head_commit = self.repo.find_commit(head_id)?;
         let head_tree = self.repo.find_tree(head_commit.tree_id())?;
 
+        let base_commit = if 0 < head_commit.parent_count() {
+            head_commit.parent(0)?
+        } else {
+            head_commit.clone()
+        };
+        let base_tree = self.repo.find_tree(base_commit.tree_id())?;
+
         let into_commit = self.repo.find_commit(into_id)?;
         let into_tree = self.repo.find_tree(into_commit.tree_id())?;
 
@@ -408,7 +433,7 @@ impl GitRepo {
             return Err(git2::Error::new(
                 git2::ErrorCode::Unmerged,
                 git2::ErrorClass::Index,
-                format!("cherry-pick conflicts:\n  {}\n", conflicts),
+                format!("squash conflicts:\n  {}\n", conflicts),
             ));
         }
         let result_id = result_index.write_tree_to(&self.repo)?;
@@ -557,9 +582,24 @@ impl GitRepo {
         self.repo.checkout_head(Some(&mut builder))?;
         Ok(())
     }
+
+    fn intern_string(&self, data: &str) -> std::rc::Rc<str> {
+        let mut interned_strings = self.interned_strings.borrow_mut();
+        if let Some(interned) = interned_strings.get(data) {
+            std::rc::Rc::clone(interned)
+        } else {
+            let interned = std::rc::Rc::from(data);
+            interned_strings.insert(std::rc::Rc::clone(&interned));
+            interned
+        }
+    }
 }
 
 impl Repo for GitRepo {
+    fn user(&self) -> Option<std::rc::Rc<str>> {
+        self.user()
+    }
+
     fn is_dirty(&self) -> bool {
         self.is_dirty()
     }
@@ -644,6 +684,7 @@ impl Repo for GitRepo {
     }
 }
 
+#[derive(Debug)]
 pub struct InMemoryRepo {
     commits: std::collections::HashMap<git2::Oid, (Option<git2::Oid>, std::rc::Rc<Commit>)>,
     branches: std::collections::HashMap<String, Branch>,
@@ -670,7 +711,7 @@ impl InMemoryRepo {
         let last_id = self
             .last_id
             .fetch_add(1, std::sync::atomic::Ordering::SeqCst);
-        let sha = format!("{:x}", last_id);
+        let sha = format!("{:040x}", last_id);
         git2::Oid::from_str(&sha).unwrap()
     }
 
@@ -697,6 +738,10 @@ impl InMemoryRepo {
         self.branches.insert(branch.name.clone(), branch);
     }
 
+    fn user(&self) -> Option<std::rc::Rc<str>> {
+        None
+    }
+
     pub fn is_dirty(&self) -> bool {
         false
     }
@@ -897,6 +942,10 @@ impl<'c> Iterator for CommitsFrom<'c> {
 }
 
 impl Repo for InMemoryRepo {
+    fn user(&self) -> Option<std::rc::Rc<str>> {
+        self.user()
+    }
+
     fn is_dirty(&self) -> bool {
         self.is_dirty()
     }
