@@ -1,10 +1,7 @@
 #![allow(clippy::collapsible_else_if)]
 
-#[macro_use]
-extern crate clap;
-
-pub mod backup;
 pub mod config;
 pub mod git;
 pub mod graph;
 pub mod log;
+pub mod stash;
